@@ -5,7 +5,7 @@
 
   This program tests the ioctl.c example file systsem.
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 
@@ -21,16 +21,15 @@
  * \include ioctl_client.c
  */
 
-#include <config.h>
-
 #include <sys/types.h>
-#include <sys/fcntl.h>
+#include <fcntl.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
 #include <errno.h>
+#include <unistd.h>
 #include "ioctl.h"
 
 const char *usage =
@@ -43,6 +42,7 @@ int main(int argc, char **argv)
 {
 	size_t size;
 	int fd;
+	int ret = 0;
 
 	if (argc < 2) {
 		fprintf(stderr, "%s", usage);
@@ -58,15 +58,19 @@ int main(int argc, char **argv)
 	if (argc == 2) {
 		if (ioctl(fd, FIOC_GET_SIZE, &size)) {
 			perror("ioctl");
-			return 1;
+			ret = 1;
+			goto out;
 		}
 		printf("%zu\n", size);
 	} else {
 		size = strtoul(argv[2], NULL, 0);
 		if (ioctl(fd, FIOC_SET_SIZE, &size)) {
 			perror("ioctl");
-			return 1;
+			ret = 1;
+			goto out;
 		}
 	}
-	return 0;
+out:
+	close(fd);
+	return ret;
 }
