@@ -10,11 +10,12 @@
   See the file COPYING.LIB.
 */
 
-#include "config.h"
+#include "fuse_config.h"
 #include "fuse_i.h"
 #include "fuse_misc.h"
 #include "fuse_opt.h"
 #include "fuse_lowlevel.h"
+#include "mount_util.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -48,6 +49,8 @@ static const struct fuse_opt fuse_helper_opts[] = {
 	FUSE_OPT_KEY("subtype=",	FUSE_OPT_KEY_KEEP),
 #endif
 	FUSE_HELPER_OPT("clone_fd",	clone_fd),
+	FUSE_HELPER_OPT("max_idle_threads=%u", max_idle_threads),
+	FUSE_HELPER_OPT("max_threads=%u", max_threads),
 	FUSE_OPT_END
 };
 
@@ -132,7 +135,11 @@ void fuse_cmdline_help(void)
 	       "    -f                     foreground operation\n"
 	       "    -s                     disable multi-threaded operation\n"
 	       "    -o clone_fd            use separate fuse device fd for each thread\n"
-	       "                           (may improve performance)\n");
+	       "                           (may improve performance)\n"
+	       "    -o max_idle_threads    the maximum number of idle worker threads\n"
+	       "                           allowed (default: -1)\n"
+	       "    -o max_threads         the maximum number of worker threads\n"
+	       "                           allowed (default: 10)\n");
 }
 
 static int fuse_helper_opt_proc(void *data, const char *arg, int key,
@@ -144,16 +151,20 @@ static int fuse_helper_opt_proc(void *data, const char *arg, int key,
 	switch (key) {
 	case FUSE_OPT_KEY_NONOPT:
 		if (!opts->mountpoint) {
-			char mountpoint[PATH_MAX];
+			if (fuse_mnt_parse_fuse_fd(arg) != -1) {
+				return fuse_opt_add_opt(&opts->mountpoint, arg);
+			}
+
+			char mountpoint[PATH_MAX] = "";
 			if (realpath(arg, mountpoint) == NULL) {
-				fprintf(stderr,
+				fuse_log(FUSE_LOG_ERR,
 					"fuse: bad mount point `%s': %s\n",
 					arg, strerror(errno));
 				return -1;
 			}
 			return fuse_opt_add_opt(&opts->mountpoint, mountpoint);
 		} else {
-			fprintf(stderr, "fuse: invalid argument `%s'\n", arg);
+			fuse_log(FUSE_LOG_ERR, "fuse: invalid argument `%s'\n", arg);
 			return -1;
 		}
 
@@ -178,7 +189,7 @@ static int add_default_subtype(const char *progname, struct fuse_args *args)
 
 	subtype_opt = (char *) malloc(strlen(basename) + 64);
 	if (subtype_opt == NULL) {
-		fprintf(stderr, "fuse: memory allocation failed\n");
+		fuse_log(FUSE_LOG_ERR, "fuse: memory allocation failed\n");
 		return -1;
 	}
 #ifdef __FreeBSD__
@@ -191,10 +202,17 @@ static int add_default_subtype(const char *progname, struct fuse_args *args)
 	return res;
 }
 
-int fuse_parse_cmdline(struct fuse_args *args,
-		       struct fuse_cmdline_opts *opts)
+int fuse_parse_cmdline_312(struct fuse_args *args,
+			   struct fuse_cmdline_opts *opts);
+FUSE_SYMVER("fuse_parse_cmdline_312", "fuse_parse_cmdline@@FUSE_3.12")
+int fuse_parse_cmdline_312(struct fuse_args *args,
+			   struct fuse_cmdline_opts *opts)
 {
 	memset(opts, 0, sizeof(struct fuse_cmdline_opts));
+
+	opts->max_idle_threads = UINT_MAX; /* new default in fuse version 3.12 */
+	opts->max_threads = 10;
+
 	if (fuse_opt_parse(args, opts, fuse_helper_opts,
 			   fuse_helper_opt_proc) == -1)
 		return -1;
@@ -210,6 +228,27 @@ int fuse_parse_cmdline(struct fuse_args *args,
 	return 0;
 }
 
+/**
+ * struct fuse_cmdline_opts got extended in libfuse-3.12
+ */
+int fuse_parse_cmdline_30(struct fuse_args *args,
+		       struct fuse_cmdline_opts *opts);
+FUSE_SYMVER("fuse_parse_cmdline_30", "fuse_parse_cmdline@FUSE_3.0")
+int fuse_parse_cmdline_30(struct fuse_args *args,
+			  struct fuse_cmdline_opts *out_opts)
+{
+	struct fuse_cmdline_opts opts;
+
+	int rc = fuse_parse_cmdline_312(args, &opts);
+	if (rc == 0) {
+		/* copy up to the size of the old pre 3.12 struct */
+		memcpy(out_opts, &opts,
+		       offsetof(struct fuse_cmdline_opts, max_idle_threads) +
+		       sizeof(opts.max_idle_threads));
+	}
+
+	return rc;
+}
 
 int fuse_daemonize(int foreground)
 {
@@ -254,7 +293,7 @@ int fuse_daemonize(int foreground)
 				close(nullfd);
 		}
 
-		/* Propagate completion of daemon initializatation */
+		/* Propagate completion of daemon initialization */
 		completed = 1;
 		(void) write(waiter[1], &completed, sizeof(completed));
 		close(waiter[0]);
@@ -265,13 +304,15 @@ int fuse_daemonize(int foreground)
 	return 0;
 }
 
-int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
-		   size_t op_size, void *user_data)
+int fuse_main_real_versioned(int argc, char *argv[],
+			     const struct fuse_operations *op, size_t op_size,
+			     struct libfuse_version *version, void *user_data)
 {
 	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
 	struct fuse *fuse;
 	struct fuse_cmdline_opts opts;
 	int res;
+	struct fuse_loop_config *loop_config = NULL;
 
 	if (fuse_parse_cmdline(&args, &opts) != 0)
 		return 1;
@@ -284,7 +325,7 @@ int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
 	}
 
 	if (opts.show_help) {
-		if(args.argv[0] != '\0')
+		if(args.argv[0][0] != '\0')
 			printf("usage: %s [options] <mountpoint>\n\n",
 			       args.argv[0]);
 		printf("FUSE options:\n");
@@ -296,40 +337,54 @@ int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
 
 	if (!opts.show_help &&
 	    !opts.mountpoint) {
-		fprintf(stderr, "error: no mountpoint specified\n");
-		res = 1;
+		fuse_log(FUSE_LOG_ERR, "error: no mountpoint specified\n");
+		res = 2;
 		goto out1;
 	}
 
-
-	fuse = fuse_new(&args, op, op_size, user_data);
+	struct fuse *_fuse_new_31(struct fuse_args *args,
+			       const struct fuse_operations *op, size_t op_size,
+			       struct libfuse_version *version,
+			       void *user_data);
+	fuse = _fuse_new_31(&args, op, op_size, version, user_data);
 	if (fuse == NULL) {
-		res = 1;
+		res = 3;
 		goto out1;
 	}
 
 	if (fuse_mount(fuse,opts.mountpoint) != 0) {
-		res = 1;
+		res = 4;
 		goto out2;
 	}
 
 	if (fuse_daemonize(opts.foreground) != 0) {
-		res = 1;
+		res = 5;
 		goto out3;
 	}
 
 	struct fuse_session *se = fuse_get_session(fuse);
 	if (fuse_set_signal_handlers(se) != 0) {
-		res = 1;
+		res = 6;
 		goto out3;
 	}
 
 	if (opts.singlethread)
 		res = fuse_loop(fuse);
-	else
-		res = fuse_loop_mt(fuse, opts.clone_fd);
+	else {
+		loop_config = fuse_loop_cfg_create();
+		if (loop_config == NULL) {
+			res = 7;
+			goto out3;
+		}
+
+		fuse_loop_cfg_set_clone_fd(loop_config, opts.clone_fd);
+
+		fuse_loop_cfg_set_idle_threads(loop_config, opts.max_idle_threads);
+		fuse_loop_cfg_set_max_threads(loop_config, opts.max_threads);
+		res = fuse_loop_mt(fuse, loop_config);
+	}
 	if (res)
-		res = 1;
+		res = 8;
 
 	fuse_remove_signal_handlers(se);
 out3:
@@ -337,11 +392,22 @@ out3:
 out2:
 	fuse_destroy(fuse);
 out1:
+	fuse_loop_cfg_destroy(loop_config);
 	free(opts.mountpoint);
 	fuse_opt_free_args(&args);
 	return res;
 }
 
+/* Not symboled, as not part of the official API */
+int fuse_main_real_30(int argc, char *argv[], const struct fuse_operations *op,
+		      size_t op_size, void *user_data);
+int fuse_main_real_30(int argc, char *argv[], const struct fuse_operations *op,
+		      size_t op_size, void *user_data)
+{
+	struct libfuse_version version = { 0 };
+	return fuse_main_real_versioned(argc, argv, op, op_size, &version,
+					user_data);
+}
 
 void fuse_apply_conn_info_opts(struct fuse_conn_info_opts *opts,
 			       struct fuse_conn_info *conn)
@@ -358,9 +424,9 @@ void fuse_apply_conn_info_opts(struct fuse_conn_info_opts *opts,
 		conn->max_readahead = opts->max_readahead;
 
 #define LL_ENABLE(cond,cap) \
-	if (cond) conn->want |= (cap)
+	if (cond) conn->want_ext |= (cap)
 #define LL_DISABLE(cond,cap) \
-	if (cond) conn->want &= ~(cap)
+	if (cond) conn->want_ext &= ~(cap)
 
 	LL_ENABLE(opts->splice_read, FUSE_CAP_SPLICE_READ);
 	LL_DISABLE(opts->no_splice_read, FUSE_CAP_SPLICE_READ);
@@ -396,7 +462,7 @@ struct fuse_conn_info_opts* fuse_parse_conn_info_opts(struct fuse_args *args)
 
 	opts = calloc(1, sizeof(struct fuse_conn_info_opts));
 	if(opts == NULL) {
-		fprintf(stderr, "calloc failed\n");
+		fuse_log(FUSE_LOG_ERR, "calloc failed\n");
 		return NULL;
 	}
 	if(fuse_opt_parse(args, opts, conn_info_opt_spec, NULL) == -1) {
@@ -405,3 +471,21 @@ struct fuse_conn_info_opts* fuse_parse_conn_info_opts(struct fuse_args *args)
 	}
 	return opts;
 }
+
+int fuse_open_channel(const char *mountpoint, const char* options)
+{
+	struct mount_opts *opts = NULL;
+	int fd = -1;
+	const char *argv[] = { "", "-o", options };
+	int argc = sizeof(argv) / sizeof(argv[0]);
+	struct fuse_args args = FUSE_ARGS_INIT(argc, (char**) argv);
+
+	opts = parse_mount_opts(&args);
+	if (opts == NULL)
+		return -1;
+
+	fd = fuse_kern_mount(mountpoint, opts);
+	destroy_mount_opts(opts);
+
+	return fd;
+}
