@@ -2,15 +2,16 @@
   FUSE: Filesystem in Userspace
   Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>
 
-  This program can be distributed under the terms of the GNU GPL.
+  This program can be distributed under the terms of the GNU GPLv2.
   See the file COPYING.
 */
 /* This program does the mounting and unmounting of FUSE filesystems */
 
-#define _GNU_SOURCE /* for clone */
-#include <config.h>
-
+#define _GNU_SOURCE /* for clone,strchrnul and close_range */
+#include "fuse_config.h"
 #include "mount_util.h"
+#include "util.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -24,36 +25,24 @@
 #include <mntent.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
-#include <sys/mount.h>
+#include <sys/param.h>
+
+#include "fuse_mount_compat.h"
+
 #include <sys/fsuid.h>
 #include <sys/socket.h>
 #include <sys/utsname.h>
 #include <sched.h>
+#include <stdbool.h>
+#include <sys/vfs.h>
+
+#ifdef HAVE_LINUX_CLOSE_RANGE_H
+#include <linux/close_range.h>
+#endif
 
 #define FUSE_COMMFD_ENV		"_FUSE_COMMFD"
 
 #define FUSE_DEV "/dev/fuse"
-#define FUSE_CONF "/etc/fuse.conf"
-
-#ifndef MS_DIRSYNC
-#define MS_DIRSYNC 128
-#endif
-#ifndef MS_REC
-#define MS_REC 16384
-#endif
-#ifndef MS_PRIVATE
-#define MS_PRIVATE (1<<18)
-#endif
-
-#ifndef UMOUNT_DETACH
-#define UMOUNT_DETACH	0x00000002	/* Just detach from the tree */
-#endif
-#ifndef UMOUNT_NOFOLLOW
-#define UMOUNT_NOFOLLOW	0x00000008	/* Don't follow symlink on umount */
-#endif
-#ifndef UMOUNT_UNUSED
-#define UMOUNT_UNUSED	0x80000000	/* Flag guaranteed to be unused */
-#endif
 
 static const char *progname;
 
@@ -62,6 +51,128 @@ static int mount_max = 1000;
 
 static int auto_unmount = 0;
 
+#ifdef GETMNTENT_NEEDS_UNESCAPING
+// Older versions of musl libc don't unescape entries in /etc/mtab
+
+// unescapes octal sequences like \040 in-place
+// That's ok, because unescaping can not extend the length of the string.
+static void unescape(char *buf) {
+	char *src = buf;
+	char *dest = buf;
+	while (1) {
+		char *next_src = strchrnul(src, '\\');
+		int offset = next_src - src;
+		memmove(dest, src, offset);
+		src = next_src;
+		dest += offset;
+
+		if(*src == '\0') {
+			*dest = *src;
+			return;
+		}
+		src++;
+
+		if('0' <= src[0] && src[0] < '2' &&
+		   '0' <= src[1] && src[1] < '8' &&
+		   '0' <= src[2] && src[2] < '8') {
+			*dest++ = (src[0] - '0') << 6
+			        | (src[1] - '0') << 3
+			        | (src[2] - '0') << 0;
+			src += 3;
+		} else if (src[0] == '\\') {
+			*dest++ = '\\';
+			src += 1;
+		} else {
+			*dest++ = '\\';
+		}
+	}
+}
+
+static struct mntent *GETMNTENT(FILE *stream)
+{
+	struct mntent *entp = getmntent(stream);
+	if(entp != NULL) {
+		unescape(entp->mnt_fsname);
+		unescape(entp->mnt_dir);
+		unescape(entp->mnt_type);
+		unescape(entp->mnt_opts);
+	}
+	return entp;
+}
+#else
+#define GETMNTENT getmntent
+#endif // GETMNTENT_NEEDS_UNESCAPING
+
+/*
+ * Take a ',' separated option string and extract "x-" options
+ */
+static int extract_x_options(const char *original, char **non_x_opts,
+			     char **x_opts)
+{
+	size_t orig_len;
+	const char *opt, *opt_end;
+
+	orig_len = strlen(original) + 1;
+
+	*non_x_opts = calloc(1, orig_len);
+	*x_opts    = calloc(1, orig_len);
+
+	size_t non_x_opts_len = orig_len;
+	size_t x_opts_len = orig_len;
+
+	if (*non_x_opts == NULL || *x_opts == NULL) {
+		fprintf(stderr, "%s: Failed to allocate %zuB.\n",
+			__func__, orig_len);
+		return -ENOMEM;
+	}
+
+	for (opt = original; opt < original + orig_len; opt = opt_end + 1) {
+		char *opt_buf;
+
+		opt_end = strchr(opt, ',');
+		if (opt_end == NULL)
+			opt_end = original + orig_len;
+
+		size_t opt_len = opt_end - opt;
+		size_t opt_len_left = orig_len - (opt - original);
+		size_t buf_len;
+		bool is_x_opts;
+
+		if (strncmp(opt, "x-", MIN(2, opt_len_left)) == 0) {
+			buf_len = x_opts_len;
+			is_x_opts = true;
+			opt_buf = *x_opts;
+		} else {
+			buf_len = non_x_opts_len;
+			is_x_opts = false;
+			opt_buf = *non_x_opts;
+		}
+
+		if (buf_len < orig_len) {
+			strncat(opt_buf, ",", 2);
+			buf_len -= 1;
+		}
+
+		/* omits ',' */
+		if ((ssize_t)(buf_len - opt_len) < 0) {
+			/* This would be a bug */
+			fprintf(stderr, "%s: no buf space left in copy, orig='%s'\n",
+				__func__, original);
+			return -EIO;
+		}
+
+		strncat(opt_buf, opt, opt_end - opt);
+		buf_len -= opt_len;
+
+		if (is_x_opts)
+			x_opts_len = buf_len;
+		else
+			non_x_opts_len = buf_len;
+	}
+
+	return 0;
+}
+
 static const char *get_user_name(void)
 {
 	struct passwd *pw = getpwuid(getuid());
@@ -168,7 +279,7 @@ static int may_unmount(const char *mnt, int quiet)
 	uidlen = sprintf(uidstr, "%u", getuid());
 
 	found = 0;
-	while ((entp = getmntent(fp)) != NULL) {
+	while ((entp = GETMNTENT(fp)) != NULL) {
 		if (!found && strcmp(entp->mnt_dir, mnt) == 0 &&
 		    (strcmp(entp->mnt_type, "fuse") == 0 ||
 		     strcmp(entp->mnt_type, "fuseblk") == 0 ||
@@ -207,6 +318,7 @@ static int may_unmount(const char *mnt, int quiet)
 
 	return 0;
 }
+#endif
 
 /*
  * Check whether the file specified in "fusermount3 -u" is really a
@@ -236,6 +348,7 @@ static int check_is_mount_child(void *p)
 	const char **a = p;
 	const char *last = a[0];
 	const char *mnt = a[1];
+	const char *type = a[2];
 	int res;
 	const char *procmounts = "/proc/mounts";
 	int found;
@@ -258,7 +371,7 @@ static int check_is_mount_child(void *p)
 	}
 
 	count = 0;
-	while (getmntent(fp) != NULL)
+	while (GETMNTENT(fp) != NULL)
 		count++;
 	endmntent(fp);
 
@@ -277,13 +390,14 @@ static int check_is_mount_child(void *p)
 	}
 
 	found = 0;
-	while ((entp = getmntent(fp)) != NULL) {
+	while ((entp = GETMNTENT(fp)) != NULL) {
 		if (count > 0) {
 			count--;
 			continue;
 		}
 		if (entp->mnt_dir[0] == '/' &&
-		    strcmp(entp->mnt_dir + 1, last) == 0) {
+		    strcmp(entp->mnt_dir + 1, last) == 0 &&
+		    (!type || strcmp(entp->mnt_type, type) == 0)) {
 			found = 1;
 			break;
 		}
@@ -316,11 +430,11 @@ static pid_t clone_newns(void *a)
 #endif
 }
 
-static int check_is_mount(const char *last, const char *mnt)
+static int check_is_mount(const char *last, const char *mnt, const char *type)
 {
 	pid_t pid, p;
 	int status;
-	const char *a[2] = { last, mnt };
+	const char *a[3] = { last, mnt, type };
 
 	pid = clone_newns((void *) a);
 	if (pid == (pid_t) -1) {
@@ -392,26 +506,13 @@ static int chdir_to_parent(char *copy, const char **lastp)
 	return 0;
 }
 
-/* Check whether the kernel supports UMOUNT_NOFOLLOW flag */
-static int umount_nofollow_support(void)
-{
-	int res = umount2("", UMOUNT_UNUSED);
-	if (res != -1 || errno != EINVAL)
-		return 0;
-
-	res = umount2("", UMOUNT_NOFOLLOW);
-	if (res != -1 || errno != ENOENT)
-		return 0;
-
-	return 1;
-}
-
+#ifndef IGNORE_MTAB
 static int unmount_fuse_locked(const char *mnt, int quiet, int lazy)
 {
 	int res;
 	char *copy;
 	const char *last;
-	int umount_flags = lazy ? UMOUNT_DETACH : 0;
+	int umount_flags = (lazy ? UMOUNT_DETACH : 0) | UMOUNT_NOFOLLOW;
 
 	if (getuid() != 0) {
 		res = may_unmount(mnt, quiet);
@@ -425,25 +526,22 @@ static int unmount_fuse_locked(const char *mnt, int quiet, int lazy)
 		return -1;
 	}
 
+	drop_privs();
 	res = chdir_to_parent(copy, &last);
-	if (res == -1)
+	if (res == -1) {
+		restore_privs();
 		goto out;
-
-	if (umount_nofollow_support()) {
-		umount_flags |= UMOUNT_NOFOLLOW;
-	} else {
-		res = check_is_mount(last, mnt);
-		if (res == -1)
-			goto out;
 	}
 
 	res = umount2(last, umount_flags);
+	restore_privs();
 	if (res == -1 && !quiet) {
 		fprintf(stderr, "%s: failed to unmount %s: %s\n",
 			progname, mnt, strerror(errno));
 	}
 
 out:
+	free(copy);
 	if (res == -1)
 		return -1;
 
@@ -478,7 +576,7 @@ static int count_fuse_fs(void)
 			strerror(errno));
 		return -1;
 	}
-	while ((entp = getmntent(fp)) != NULL) {
+	while ((entp = GETMNTENT(fp)) != NULL) {
 		if (strcmp(entp->mnt_type, "fuse") == 0 ||
 		    strncmp(entp->mnt_type, "fuse.", 5) == 0)
 			count ++;
@@ -489,7 +587,7 @@ static int count_fuse_fs(void)
 
 
 #else /* IGNORE_MTAB */
-static int count_fuse_fs()
+static int count_fuse_fs(void)
 {
 	return 0;
 }
@@ -506,6 +604,7 @@ static int add_mount(const char *source, const char *mnt, const char *type,
 
 static int unmount_fuse(const char *mnt, int quiet, int lazy)
 {
+	(void) quiet;
 	return fuse_mnt_umount(progname, mnt, mnt, lazy);
 }
 #endif /* IGNORE_MTAB */
@@ -563,10 +662,19 @@ static void read_conf(void)
 			fprintf(stderr, "%s: reading %s: missing newline at end of file\n", progname, FUSE_CONF);
 
 		}
+		if (ferror(fp)) {
+			fprintf(stderr, "%s: reading %s: read failed\n", progname, FUSE_CONF);
+			exit(1);
+		}
 		fclose(fp);
 	} else if (errno != ENOENT) {
+		bool fatal = (errno != EACCES && errno != ELOOP &&
+			      errno != ENAMETOOLONG && errno != ENOTDIR &&
+			      errno != EOVERFLOW);
 		fprintf(stderr, "%s: failed to open %s: %s\n",
 			progname, FUSE_CONF, strerror(errno));
+		if (fatal)
+			exit(1);
 	}
 }
 
@@ -598,7 +706,17 @@ static struct mount_flags mount_flags[] = {
 	{"sync",    MS_SYNCHRONOUS, 1, 1},
 	{"atime",   MS_NOATIME,	    0, 1},
 	{"noatime", MS_NOATIME,	    1, 1},
+	{"diratime",        MS_NODIRATIME,  0, 1},
+	{"nodiratime",      MS_NODIRATIME,  1, 1},
+	{"lazytime",        MS_LAZYTIME,    1, 1},
+	{"nolazytime",      MS_LAZYTIME,    0, 1},
+	{"relatime",        MS_RELATIME,    1, 1},
+	{"norelatime",      MS_RELATIME,    0, 1},
+	{"strictatime",     MS_STRICTATIME, 1, 1},
+	{"nostrictatime",   MS_STRICTATIME, 0, 1},
 	{"dirsync", MS_DIRSYNC,	    1, 1},
+	{"symfollow",       MS_NOSYMFOLLOW, 0, 1},
+	{"nosymfollow",     MS_NOSYMFOLLOW, 1, 1},
 	{NULL,	    0,		    0, 0}
 };
 
@@ -710,7 +828,24 @@ static int get_string_opt(const char *s, unsigned len, const char *opt,
 	return 1;
 }
 
-static int do_mount(const char *mnt, char **typep, mode_t rootmode,
+/* The kernel silently truncates the "data" argument to PAGE_SIZE-1 characters.
+ * This can be dangerous if it e.g. truncates the option "group_id=1000" to
+ * "group_id=1".
+ * This wrapper detects this case and bails out with an error.
+ */
+static int mount_notrunc(const char *source, const char *target,
+			 const char *filesystemtype, unsigned long mountflags,
+			 const char *data) {
+	if (strlen(data) > sysconf(_SC_PAGESIZE) - 1) {
+		fprintf(stderr, "%s: mount options too long\n", progname);
+		errno = EINVAL;
+		return -1;
+	}
+	return mount(source, target, filesystemtype, mountflags, data);
+}
+
+
+static int do_mount(const char *mnt, const char **typep, mode_t rootmode,
 		    int fd, const char *opts, const char *dev, char **sourcep,
 		    char **mnt_optsp)
 {
@@ -736,8 +871,10 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 		unsigned len;
 		const char *fsname_str = "fsname=";
 		const char *subtype_str = "subtype=";
+		bool escape_ok = begins_with(s, fsname_str) ||
+				 begins_with(s, subtype_str);
 		for (len = 0; s[len]; len++) {
-			if (s[len] == '\\' && s[len + 1])
+			if (escape_ok && s[len] == '\\' && s[len + 1])
 				len++;
 			else if (s[len] == ',')
 				break;
@@ -758,7 +895,8 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 			blkdev = 1;
 		} else if (opt_eq(s, len, "auto_unmount")) {
 			auto_unmount = 1;
-		} else if (!begins_with(s, "fd=") &&
+		} else if (!opt_eq(s, len, "nonempty") &&
+			   !begins_with(s, "fd=") &&
 			   !begins_with(s, "rootmode=") &&
 			   !begins_with(s, "user_id=") &&
 			   !begins_with(s, "group_id=")) {
@@ -789,10 +927,16 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 						flags |= flag;
 					else
 						flags  &= ~flag;
-				} else {
+				} else if (opt_eq(s, len, "default_permissions") ||
+					   opt_eq(s, len, "allow_other") ||
+					   begins_with(s, "max_read=") ||
+					   begins_with(s, "blksize=")) {
 					memcpy(d, s, len);
 					d += len;
 					*d++ = ',';
+				} else {
+					fprintf(stderr, "%s: unknown option '%.*s'\n", progname, len, s);
+					exit(1);
 				}
 			}
 		}
@@ -827,7 +971,7 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 	else
 		strcpy(source, subtype ? subtype : dev);
 
-	res = mount(source, mnt, type, flags, optbuf);
+	res = mount_notrunc(source, mnt, type, flags, optbuf);
 	if (res == -1 && errno == ENODEV && subtype) {
 		/* Probably missing subtype support */
 		strcpy(type, blkdev ? "fuseblk" : "fuse");
@@ -838,13 +982,13 @@ static int do_mount(const char *mnt, char **typep, mode_t rootmode,
 			strcpy(source, type);
 		}
 
-		res = mount(source, mnt, type, flags, optbuf);
+		res = mount_notrunc(source, mnt, type, flags, optbuf);
 	}
 	if (res == -1 && errno == EINVAL) {
 		/* It could be an old version not supporting group_id */
 		sprintf(d, "fd=%i,rootmode=%o,user_id=%u",
 			fd, rootmode, getuid());
-		res = mount(source, mnt, type, flags, optbuf);
+		res = mount_notrunc(source, mnt, type, flags, optbuf);
 	}
 	if (res == -1) {
 		int errno_save = errno;
@@ -879,6 +1023,8 @@ static int check_perm(const char **mntp, struct stat *stbuf, int *mountpoint_fd)
 	int res;
 	const char *mnt = *mntp;
 	const char *origmnt = mnt;
+	struct statfs fs_buf;
+	size_t i;
 
 	res = lstat(mnt, stbuf);
 	if (res == -1) {
@@ -951,8 +1097,70 @@ static int check_perm(const char **mntp, struct stat *stbuf, int *mountpoint_fd)
 		return -1;
 	}
 
+	/* Do not permit mounting over anything in procfs - it has a couple
+	 * places to which we have "write access" without being supposed to be
+	 * able to just put anything we want there.
+	 * Luckily, without allow_other, we can't get other users to actually
+	 * use any fake information we try to put there anyway.
+	 * Use a whitelist to be safe. */
+	if (statfs(*mntp, &fs_buf)) {
+		fprintf(stderr, "%s: failed to access mountpoint %s: %s\n",
+			progname, mnt, strerror(errno));
+		return -1;
+	}
 
-	return 0;
+	/* Define permitted filesystems for the mount target. This was
+	 * originally the same list as used by the ecryptfs mount helper
+	 * (https://bazaar.launchpad.net/~ecryptfs/ecryptfs/trunk/view/head:/src/utils/mount.ecryptfs_private.c#L225)
+	 * but got expanded as we found more filesystems that needed to be
+	 * overlaid. */
+	typeof(fs_buf.f_type) f_type_whitelist[] = {
+		0x61756673 /* AUFS_SUPER_MAGIC */,
+		0x00000187 /* AUTOFS_SUPER_MAGIC */,
+		0xCA451A4E /* BCACHEFS_STATFS_MAGIC */,
+		0x9123683E /* BTRFS_SUPER_MAGIC */,
+		0x00C36400 /* CEPH_SUPER_MAGIC */,
+		0xFF534D42 /* CIFS_MAGIC_NUMBER */,
+		0x0000F15F /* ECRYPTFS_SUPER_MAGIC */,
+		0X2011BAB0 /* EXFAT_SUPER_MAGIC */,
+		0x0000EF53 /* EXT[234]_SUPER_MAGIC */,
+		0xF2F52010 /* F2FS_SUPER_MAGIC */,
+		0x65735546 /* FUSE_SUPER_MAGIC */,
+		0x01161970 /* GFS2_MAGIC */,
+		0x47504653 /* GPFS_SUPER_MAGIC */,
+		0x0000482b /* HFSPLUS_SUPER_MAGIC */,
+		0x000072B6 /* JFFS2_SUPER_MAGIC */,
+		0x3153464A /* JFS_SUPER_MAGIC */,
+		0x0BD00BD0 /* LL_SUPER_MAGIC */,
+		0X00004D44 /* MSDOS_SUPER_MAGIC */,
+		0x0000564C /* NCP_SUPER_MAGIC */,
+		0x00006969 /* NFS_SUPER_MAGIC */,
+		0x00003434 /* NILFS_SUPER_MAGIC */,
+		0x5346544E /* NTFS_SB_MAGIC */,
+		0x7366746E /* NTFS3_SUPER_MAGIC */,
+		0x5346414f /* OPENAFS_SUPER_MAGIC */,
+		0x794C7630 /* OVERLAYFS_SUPER_MAGIC */,
+		0xAAD7AAEA /* PANFS_SUPER_MAGIC */,
+		0x52654973 /* REISERFS_SUPER_MAGIC */,
+		0xFE534D42 /* SMB2_SUPER_MAGIC */,
+		0x73717368 /* SQUASHFS_MAGIC */,
+		0x01021994 /* TMPFS_MAGIC */,
+		0x24051905 /* UBIFS_SUPER_MAGIC */,
+#if __SIZEOF_LONG__ > 4
+		0x736675005346544e /* UFSD */,
+#endif
+		0x58465342 /* XFS_SB_MAGIC */,
+		0x2FC12FC1 /* ZFS_SUPER_MAGIC */,
+		0x858458f6 /* RAMFS_MAGIC */,
+	};
+	for (i = 0; i < sizeof(f_type_whitelist)/sizeof(f_type_whitelist[0]); i++) {
+		if (f_type_whitelist[i] == fs_buf.f_type)
+			return 0;
+	}
+
+	fprintf(stderr, "%s: mounting over filesystem type %#010lx is forbidden\n",
+		progname, (unsigned long)fs_buf.f_type);
+	return -1;
 }
 
 static int try_open(const char *dev, char **devp, int silent)
@@ -1000,17 +1208,18 @@ static int open_fuse_device(char **devp)
 }
 
 
-static int mount_fuse(const char *mnt, const char *opts)
+static int mount_fuse(const char *mnt, const char *opts, const char **type)
 {
 	int res;
 	int fd;
 	char *dev;
 	struct stat stbuf;
-	char *type = NULL;
 	char *source = NULL;
 	char *mnt_opts = NULL;
 	const char *real_mnt = mnt;
 	int mountpoint_fd = -1;
+	char *do_mount_opts = NULL;
+	char *x_opts = NULL;
 
 	fd = open_fuse_device(&dev);
 	if (fd == -1)
@@ -1027,11 +1236,16 @@ static int mount_fuse(const char *mnt, const char *opts)
 		}
 	}
 
+	// Extract any options starting with "x-"
+	res= extract_x_options(opts, &do_mount_opts, &x_opts);
+	if (res)
+		goto fail_close_fd;
+
 	res = check_perm(&real_mnt, &stbuf, &mountpoint_fd);
 	restore_privs();
 	if (res != -1)
-		res = do_mount(real_mnt, &type, stbuf.st_mode & S_IFMT,
-			       fd, opts, dev, &source, &mnt_opts);
+		res = do_mount(real_mnt, type, stbuf.st_mode & S_IFMT,
+			       fd, do_mount_opts, dev, &source, &mnt_opts);
 
 	if (mountpoint_fd != -1)
 		close(mountpoint_fd);
@@ -1046,7 +1260,29 @@ static int mount_fuse(const char *mnt, const char *opts)
 	}
 
 	if (geteuid() == 0) {
-		res = add_mount(source, mnt, type, mnt_opts);
+		if (x_opts && strlen(x_opts) > 0) {
+			/*
+			 * Add back the options starting with "x-" to opts from
+			 * do_mount. +2 for ',' and '\0'
+			 */
+			size_t mnt_opts_len = strlen(mnt_opts);
+			size_t x_mnt_opts_len =  mnt_opts_len+
+						 strlen(x_opts) + 2;
+			char *x_mnt_opts = calloc(1, x_mnt_opts_len);
+
+			if (mnt_opts_len) {
+				strcpy(x_mnt_opts, mnt_opts);
+				strncat(x_mnt_opts, ",", 2);
+			}
+
+			strncat(x_mnt_opts, x_opts,
+				x_mnt_opts_len - mnt_opts_len - 2);
+
+			free(mnt_opts);
+			mnt_opts = x_mnt_opts;
+		}
+
+		res = add_mount(source, mnt, *type, mnt_opts);
 		if (res == -1) {
 			/* Can't clean up mount in a non-racy way */
 			goto fail_close_fd;
@@ -1055,9 +1291,10 @@ static int mount_fuse(const char *mnt, const char *opts)
 
 out_free:
 	free(source);
-	free(type);
 	free(mnt_opts);
 	free(dev);
+	free(x_opts);
+	free(do_mount_opts);
 
 	return fd;
 
@@ -1103,13 +1340,109 @@ static int send_fd(int sock_fd, int fd)
 	return 0;
 }
 
+/* Helper for should_auto_unmount
+ *
+ * fusermount typically has the s-bit set - initial open of `mnt` was as root
+ * and got EACCESS as 'allow_other' was not specified.
+ * Try opening `mnt` again with uid and guid of the calling process.
+ */
+static int recheck_ENOTCONN_as_owner(const char *mnt)
+{
+	int pid = fork();
+	if(pid == -1) {
+		perror("fuse: recheck_ENOTCONN_as_owner can't fork");
+		_exit(EXIT_FAILURE);
+	} else if(pid == 0) {
+		uid_t uid = getuid();
+		gid_t gid = getgid();
+		if(setresgid(gid, gid, gid) == -1) {
+			perror("fuse: can't set resgid");
+			_exit(EXIT_FAILURE);
+		}
+		if(setresuid(uid, uid, uid) == -1) {
+			perror("fuse: can't set resuid");
+			_exit(EXIT_FAILURE);
+		}
+
+		int fd = open(mnt, O_RDONLY);
+		if(fd == -1 && errno == ENOTCONN)
+			_exit(EXIT_SUCCESS);
+		else
+			_exit(EXIT_FAILURE);
+	} else {
+		int status;
+		int res = waitpid(pid, &status, 0);
+		if (res == -1) {
+			perror("fuse: waiting for child failed");
+			_exit(EXIT_FAILURE);
+		}
+		return WIFEXITED(status) && WEXITSTATUS(status) == EXIT_SUCCESS;
+	}
+}
+
+/* The parent fuse process has died: decide whether to auto_unmount.
+ *
+ * In the normal case (umount or fusermount -u), the filesystem
+ * has already been unmounted. If we simply unmount again we can
+ * cause problems with stacked mounts (e.g. autofs).
+ *
+ * So we unmount here only in abnormal case where fuse process has
+ * died without unmount happening. To detect this, we first look in
+ * the mount table to make sure the mountpoint is still mounted and
+ * has proper type. If so, we then see if opening the mount dir is
+ * returning 'Transport endpoint is not connected'.
+ *
+ * The order of these is important, because if autofs is in use,
+ * opening the dir to check for ENOTCONN will cause a new mount
+ * in the normal case where filesystem has been unmounted cleanly.
+ */
+static int should_auto_unmount(const char *mnt, const char *type)
+{
+	char *copy;
+	const char *last;
+	int result = 0;
+	int fd;
+
+	copy = strdup(mnt);
+	if (copy == NULL) {
+	fprintf(stderr, "%s: failed to allocate memory\n", progname);
+		return 0;
+	}
+
+	if (chdir_to_parent(copy, &last) == -1)
+		goto out;
+	if (check_is_mount(last, mnt, type) == -1)
+		goto out;
+
+	fd = open(mnt, O_RDONLY);
+
+	if (fd != -1) {
+		close(fd);
+	} else {
+		switch(errno) {
+		case ENOTCONN:
+			result = 1;
+			break;
+		case EACCES:
+			result = recheck_ENOTCONN_as_owner(mnt);
+			break;
+		default:
+			result = 0;
+			break;
+		}
+	}
+out:
+	free(copy);
+	return result;
+}
+
 static void usage(void)
 {
 	printf("%s: [options] mountpoint\n"
 	       "Options:\n"
 	       " -h		    print help\n"
 	       " -V		    print version\n"
-	       " -o opt[,opt...]   mount options\n"
+	       " -o opt[,opt...]    mount options\n"
 	       " -u		    unmount\n"
 	       " -q		    quiet\n"
 	       " -z		    lazy unmount\n",
@@ -1123,6 +1456,65 @@ static void show_version(void)
 	exit(0);
 }
 
+static void close_range_loop(int min_fd, int max_fd, int cfd)
+{
+	for (int fd = min_fd; fd <= max_fd; fd++)
+		if (fd != cfd)
+			close(fd);
+}
+
+/*
+ * Close all inherited fds that are not needed
+ * Ideally these wouldn't come up at all, applications should better
+ * use FD_CLOEXEC / O_CLOEXEC
+ */
+static int close_inherited_fds(int cfd)
+{
+	int rc = -1;
+	int nullfd;
+
+	/* We can't even report an error */
+	if (cfd <= STDERR_FILENO)
+		return -EINVAL;
+
+#ifdef HAVE_LINUX_CLOSE_RANGE_H
+	if (cfd < STDERR_FILENO + 2) {
+		close_range_loop(STDERR_FILENO + 1, cfd - 1, cfd);
+	} else {
+		rc = close_range(STDERR_FILENO + 1, cfd - 1, 0);
+		if (rc < 0)
+			goto fallback;
+	}
+
+	/* Close high range */
+	rc = close_range(cfd + 1, ~0U, 0);
+#else
+	goto fallback; /* make use of fallback to avoid compiler warnings */
+#endif
+
+fallback:
+	if (rc < 0) {
+		int max_fd = sysconf(_SC_OPEN_MAX) - 1;
+
+		close_range_loop(STDERR_FILENO + 1, max_fd, cfd);
+	}
+
+	nullfd = open("/dev/null", O_RDWR);
+	if (nullfd < 0) {
+		perror("fusermount: cannot open /dev/null");
+		return -errno;
+	}
+
+	/* Redirect stdin, stdout, stderr to /dev/null */
+	dup2(nullfd, STDIN_FILENO);
+	dup2(nullfd, STDOUT_FILENO);
+	dup2(nullfd, STDERR_FILENO);
+	if (nullfd > STDERR_FILENO)
+		close(nullfd);
+
+	return 0;
+}
+
 int main(int argc, char *argv[])
 {
 	sigset_t sigset;
@@ -1134,9 +1526,11 @@ int main(int argc, char *argv[])
 	static int unmount = 0;
 	static int lazy = 0;
 	static int quiet = 0;
-	char *commfd;
-	int cfd;
+	char *commfd = NULL;
+	long cfd;
 	const char *opts = "";
+	const char *type = NULL;
+	int setup_auto_unmount_only = 0;
 
 	static const struct option long_opts[] = {
 		{"unmount", no_argument, NULL, 'u'},
@@ -1144,9 +1538,14 @@ int main(int argc, char *argv[])
 		{"quiet",   no_argument, NULL, 'q'},
 		{"help",    no_argument, NULL, 'h'},
 		{"version", no_argument, NULL, 'V'},
+		{"options", required_argument, NULL, 'o'},
+		// Note: auto-unmount and comm-fd don't have short versions.
+		// They'ne meant for internal use by mount.c
+		{"auto-unmount", no_argument, NULL, 'U'},
+		{"comm-fd", required_argument, NULL, 'c'},
 		{0, 0, 0, 0}};
 
-	progname = strdup(argv[0]);
+	progname = strdup(argc > 0 ? argv[0] : "fusermount");
 	if (progname == NULL) {
 		fprintf(stderr, "%s: failed to allocate memory\n", argv[0]);
 		exit(1);
@@ -1170,7 +1569,14 @@ int main(int argc, char *argv[])
 		case 'u':
 			unmount = 1;
 			break;
-
+		case 'U':
+			unmount = 1;
+			auto_unmount = 1;
+			setup_auto_unmount_only = 1;
+			break;
+		case 'c':
+			commfd = optarg;
+			break;
 		case 'z':
 			lazy = 1;
 			break;
@@ -1214,35 +1620,67 @@ int main(int argc, char *argv[])
 		exit(1);
 
 	umask(033);
-	if (unmount)
+	if (!setup_auto_unmount_only && unmount)
 		goto do_unmount;
 
-	commfd = getenv(FUSE_COMMFD_ENV);
+	if(commfd == NULL)
+		commfd = getenv(FUSE_COMMFD_ENV);
 	if (commfd == NULL) {
 		fprintf(stderr, "%s: old style mounting not supported\n",
 			progname);
 		goto err_out;
 	}
 
-	fd = mount_fuse(mnt, opts);
+	res = libfuse_strtol(commfd, &cfd);
+	if (res) {
+		fprintf(stderr,
+			"%s: invalid _FUSE_COMMFD: %s\n",
+			progname, commfd);
+		goto err_out;
+
+	}
+
+	{
+		struct stat statbuf;
+		fstat(cfd, &statbuf);
+		if(!S_ISSOCK(statbuf.st_mode)) {
+			fprintf(stderr,
+				"%s: file descriptor %li is not a socket, can't send fuse fd\n",
+				progname, cfd);
+			goto err_out;
+		}
+	}
+
+	if (setup_auto_unmount_only)
+		goto wait_for_auto_unmount;
+
+	fd = mount_fuse(mnt, opts, &type);
 	if (fd == -1)
 		goto err_out;
 
-	cfd = atoi(commfd);
 	res = send_fd(cfd, fd);
-	if (res == -1)
+	if (res != 0) {
+		umount2(mnt, MNT_DETACH); /* lazy umount */
 		goto err_out;
+	}
 	close(fd);
 
 	if (!auto_unmount) {
 		free(mnt);
+		free((void*) type);
 		return 0;
 	}
 
+wait_for_auto_unmount:
 	/* Become a daemon and wait for the parent to exit or die.
 	   ie For the control socket to get closed.
-	   btw We don't want to use daemon() function here because
+	   Btw, we don't want to use daemon() function here because
 	   it forks and messes with the file descriptors. */
+
+	res = close_inherited_fds(cfd);
+	if (res < 0)
+		exit(EXIT_FAILURE);
+
 	setsid();
 	res = chdir("/");
 	if (res == -1) {
@@ -1269,6 +1707,10 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (!should_auto_unmount(mnt, type)) {
+		goto success_out;
+	}
+
 do_unmount:
 	if (geteuid() == 0)
 		res = unmount_fuse(mnt, quiet, lazy);
@@ -1281,9 +1723,14 @@ do_unmount:
 	}
 	if (res == -1)
 		goto err_out;
+
+success_out:
+	free((void*) type);
+	free(mnt);
 	return 0;
 
 err_out:
+	free((void*) type);
 	free(mnt);
 	exit(1);
 }
